import { identity, tuple } from 'fp-ts/lib/function';
import { fromEither, fromNullable, none, some, option, isNone } from 'fp-ts/lib/Option';
import { filter, isEmpty } from 'fp-ts/lib/Record';
import { pipe } from 'fp-ts/lib/pipeable';
import { string, Type, Int, failure, success } from 'io-ts';
import { stringify } from 'querystring';
import { parse as parseUrl } from 'url';
import { either } from 'fp-ts/lib/Either';
/**
 * @since 0.4.0
 */
export class Route {
    constructor(parts, query) {
        this.parts = parts;
        this.query = query;
    }
    toString(encode = true) {
        const nonUndefinedQuery = pipe(this.query, filter(value => value !== undefined));
        const qs = stringify(nonUndefinedQuery);
        const parts = encode ? this.parts.map(encodeURIComponent) : this.parts;
        return '/' + parts.join('/') + (qs ? '?' + qs : '');
    }
}
Route.empty = new Route([], {});
Route.isEmpty = (r) => r.parts.length === 0 && isEmpty(r.query);
Route.parse = (s, decode = true) => {
    const route = parseUrl(s, true);
    const oparts = option.map(fromNullable(route.pathname), s => {
        const r = s.split('/').filter(Boolean);
        return decode ? r.map(decodeURIComponent) : r;
    });
    const parts = isNone(oparts) ? [] : oparts.value;
    return new Route(parts, route.query);
};
const assign = (a) => (b) => Object.assign({}, a, b);
/**
 * @since 0.4.0
 */
export class Parser {
    constructor(run) {
        this.run = run;
    }
    map(f) {
        return this.chain(a => Parser.of(f(a))); // <= derived
    }
    ap(fab) {
        return fab.chain(f => this.map(f)); // <= derived
    }
    chain(f) {
        return new Parser(r => option.chain(this.run(r), ([a, r2]) => f(a).run(r2)));
    }
    alt(that) {
        return new Parser(r => {
            const oar = this.run(r);
            return isNone(oar) ? that.run(r) : oar;
        });
    }
    /** A mapped Monoidal.mult */
    then(that) {
        return that.ap(this.map(assign));
    }
}
Parser.of = (a) => new Parser(s => some(tuple(a, s)));
/**
 * @since 0.4.0
 */
export function zero() {
    return new Parser(() => none);
}
/**
 * @since 0.4.0
 */
export function parse(parser, r, a) {
    const oa = option.map(parser.run(r), ([a]) => a);
    return isNone(oa) ? a : oa.value;
}
/**
 * @since 0.4.0
 */
export function format(formatter, a, encode = true) {
    return formatter.run(Route.empty, a).toString(encode);
}
/**
 * @since 0.4.0
 */
export class Formatter {
    constructor(run) {
        this.run = run;
    }
    contramap(f) {
        return new Formatter((r, b) => this.run(r, f(b)));
    }
    then(that) {
        return new Formatter((r, ab) => that.run(this.run(r, ab), ab));
    }
}
/**
 * @since 0.4.0
 */
export class Match {
    constructor(parser, formatter) {
        this.parser = parser;
        this.formatter = formatter;
    }
    imap(f, g) {
        return new Match(this.parser.map(f), this.formatter.contramap(g));
    }
    then(that) {
        const p = this.parser.then(that.parser);
        const f = this.formatter.then(that.formatter);
        return new Match(p, f);
    }
}
const singleton = (k, v) => ({ [k]: v });
/**
 * `succeed` matches everything but consumes nothing
 * @since 0.4.0
 */
export function succeed(a) {
    return new Match(new Parser(r => some(tuple(a, r))), new Formatter(identity));
}
/**
 * `end` matches the end of a route
 * @since 0.4.0
 */
export const end = new Match(new Parser(r => (Route.isEmpty(r) ? some(tuple({}, r)) : none)), new Formatter(identity));
/**
 * `type` matches any io-ts type path component
 * @since 0.4.0
 */
export function type(k, type) {
    return new Match(new Parser(r => {
        if (r.parts.length === 0) {
            return none;
        }
        else {
            const head = r.parts[0];
            const tail = r.parts.slice(1);
            return option.map(fromEither(type.decode(head)), a => tuple(singleton(k, a), new Route(tail, r.query)));
        }
    }), new Formatter((r, o) => new Route(r.parts.concat(type.encode(o[k])), r.query)));
}
/**
 * `str` matches any string path component
 * @since 0.4.0
 */
export function str(k) {
    return type(k, string);
}
/**
 * @internal
 */
export const IntegerFromString = new Type('IntegerFromString', (u) => Int.is(u), (u, c) => either.chain(string.validate(u, c), s => {
    const n = +s;
    return isNaN(n) || !Number.isInteger(n) ? failure(s, c) : success(n);
}), String);
/**
 * `int` matches any integer path component
 * @since 0.4.0
 */
export function int(k) {
    return type(k, IntegerFromString);
}
/**
 * `lit(x)` will match exactly the path component `x`
 * For example, `lit('x')` matches `/x`
 * @since 0.4.0
 */
export function lit(literal) {
    return new Match(new Parser(r => {
        if (r.parts.length === 0) {
            return none;
        }
        else {
            const head = r.parts[0];
            const tail = r.parts.slice(1);
            return head === literal ? some(tuple({}, new Route(tail, r.query))) : none;
        }
    }), new Formatter((r, n) => new Route(r.parts.concat(literal), r.query)));
}
/**
 * @since 0.4.0
 */
export function query(type) {
    return new Match(new Parser(r => option.map(fromEither(type.decode(r.query)), query => tuple(query, new Route(r.parts, {})))), new Formatter((r, query) => new Route(r.parts, type.encode(query))));
}
//# sourceMappingURL=index.js.map